"""Script entrypoints"""
import logging
import os
import sys

from ee_spf_flattener.config import config
from ee_spf_flattener.crawler import spf2ips
from ee_spf_flattener.formatting import compute_spf_records, format_spf_record, ips_to_spf_strings

LOG = logging.getLogger(__name__)


def update_updater_statusfile():
    """Update updater status file"""
    statusfile = config.get("update.statusfile")
    if not statusfile:
        return
    LOG.info("Update updater status file (%s)", statusfile)
    try:
        os.utime(statusfile, None)
    except FileNotFoundError:
        # pylint: disable=consider-using-with
        open(statusfile, "a", encoding="utf-8").close()


def main():
    """Script entrypoint"""
    # Parse arguments
    subcommand = config.parse_args()
    if not subcommand:
        config.error("You must provide the command as first positional argument.")

    domain = config.get("flattening.domain", required=True)
    ips = spf2ips(config.get("flattening.input_records", required=True), domain)
    LOG.debug("%d IP addresses found: %s", len(ips), ", ".join(map(str, ips)))

    if not ips:
        LOG.fatal("No SPF IP addresses found from specified input records")
        return 1

    ips = ips_to_spf_strings(ips)
    LOG.debug("%d IP addresses found after CIDR based simplification: %s", len(ips), ", ".join(ips))

    if subcommand == "update":
        updater = config.get("update.updater")
        updater.initialize()

        current_ips = spf2ips(
            [{"name": config.get("flattening.spf_record_name_format").format(""), "type": "txt"}],
            domain,
        )
        LOG.debug(
            "%d current IP addresses found: %s", len(current_ips), ", ".join(map(str, current_ips))
        )
        current_ips = ips_to_spf_strings(current_ips)
        LOG.debug(
            "%d IP addresses found after CIDR based simplification: %s",
            len(current_ips),
            ", ".join(current_ips),
        )
        if ips == current_ips:
            LOG.info("No change, flatten SPF record is up-to-date.")
            if not config.get("update.force"):
                update_updater_statusfile()
                sys.exit(0)
            LOG.info("Force NS update mode, continue")
        else:
            LOG.info(
                "Change detected in flatten SPF record:\n- added: %s\n- removed: %s",
                ", ".join(map(str, set(ips).difference(set(current_ips)))),
                ", ".join(map(str, set(current_ips).difference(set(ips)))),
            )

    spf_records = compute_spf_records(ips)
    LOG.debug(
        "%d SPF records computed:\n%s",
        len(spf_records),
        "\n".join([f"- {name}: '{value}'" for name, value in spf_records.items()]),
    )

    if not config.get("flattening.dont_warn_about_spf_record_count") and len(
        spf_records
    ) > config.get("flattening.max_spf_records"):
        LOG.warning(
            "Even with flattening, you have more than %d SPF entries",
            config.get("flattening.max_spf_records"),
        )

    if subcommand == "dump":
        print(
            "\n".join(
                [
                    format_spf_record(name, value, ttl=config.get("flattening.ttl"))
                    for name, value in spf_records.items()
                ]
            )
        )
        return 0

    if subcommand == "update":
        if config.just_try:
            LOG.info("Just-try mode: do not update NS server")
            return 0
        result = updater.update_spf_records(spf_records)
        # Handle status file
        if result:
            update_updater_statusfile()

        return 0 if result else 1

    return 1
