"""Configuration and CLI arguments handler"""

import argparse
import atexit
import importlib.util
import logging
import os
import pkgutil
import re
import sys
from urllib.parse import urlparse

import yaml
from dns.resolver import Resolver
from netaddr import IPAddress

from ee_spf_flattener.const import (
    CRAWLER_SUPPORTED_RECORD_TYPES,
    DEFAULT_GANDI_API_URL,
    DEFAULT_GANDI_AUTH_API_URL,
    DEFAULT_GANDI_WARNING_API_KEY_EXPIRATION_DELAY,
    DEFAULT_MAX_SPF_RECORDS,
    DEFAULT_NSUPDATE_KEY_ALGORITHM,
    DEFAULT_SPF_ALL_POLICY,
    DEFAULT_SPF_RECORD_MAX_LENGTH,
    DEFAULT_SPF_RECORD_NAME_FORMAT,
    DEFAULT_TIMEOUT,
    DEFAULT_TTL,
    DEFAULT_UPDATER,
    SPF_ALL_POLICIES,
)

LOG = logging.getLogger(__name__)


def loglevel(value):
    """ArgumentParser for loglevel type"""
    value = value.upper()
    if value in logging._nameToLevel:  # pylint: disable=protected-access
        return value
    raise ValueError


def input_records_arg(value):
    """ArgumentParser input record type"""
    records = []
    for input_record in value.split(","):
        input_record = input_record.strip().split(":")
        if len(input_record) not in [1, 2]:
            raise ValueError()
        records.append(
            {
                "name": input_record[0],
                "type": input_record[1].lower() if len(input_record) == 2 else "txt",
            }
        )
    return records


def input_records_config_value(values):
    """Config value for input record type"""
    if not isinstance(values, list):
        raise ValueError()
    records = []
    for value in values:
        if not isinstance(value, dict):
            raise ValueError()
        try:
            input_record = {
                "name": value["name"],
                "type": value.get("type", "txt").lower(),
            }
        except KeyError as err:
            raise ValueError from err
        if input_record["type"] not in CRAWLER_SUPPORTED_RECORD_TYPES:
            raise ValueError
        records.append(input_record)
    return records


def dns_resolvers(value):
    """ArgumentParser input record type"""
    dns_resolver_ips = []
    for ip in value if isinstance(value, list) else value.split(","):
        ip = ip.strip()
        try:
            IPAddress(ip)
        except Exception as err:
            raise ValueError() from err
        dns_resolver_ips.append(ip)
    return dns_resolver_ips


def spf_record_name_format(value):
    """ArgumentParser SPF record name format type"""
    try:
        assert "{" not in re.sub(r"\$\{[^\}]+\}", "", value).format(1)
        return value
    except Exception as err:
        raise ValueError() from err


def spf_all_policy(value):
    """ArgumentParser SPF all policy type"""
    if value in SPF_ALL_POLICIES:
        return value
    raise ValueError()


def list_updaters():
    """List valid SPF records updaters"""
    return [
        module.name
        for module in pkgutil.iter_modules(
            [os.path.dirname(importlib.util.find_spec("ee_spf_flattener.updater").origin)]
        )
    ]


def updater(value):
    """ArgumentParser updater choice"""
    try:
        module = importlib.import_module(f"ee_spf_flattener.updater.{value}")
        return module.Updater()
    except ModuleNotFoundError as err:
        raise ValueError() from err


def url(value):
    """ArgumentParser type for URL"""
    urlparse(value)
    return value[0:-1] if value.endswith("/") else value


class Config:
    """Class to manage configuration file and CLI arguments"""

    _args_parser = None
    args = None
    path = None
    config_data = {}
    config_path = None

    config_types = {
        "logging": {
            "level": loglevel,
            "logfile": str,
            "logfile_format": str,
            "logfile_level": loglevel,
            "console": bool,
            "console_format": str,
            "console_level": loglevel,
            "console_force_stderr": bool,
        },
        "flattening": {
            "domain": str,
            "input_records": input_records_config_value,
            "name_servers": dns_resolvers,
            "spf_all_policy": spf_all_policy,
            "spf_record_name_format": spf_record_name_format,
            "spf_record_max_length": int,
            "max_spf_records": int,
            "dont_warn_about_spf_record_count": bool,
            "ttl": int,
        },
        "update": {
            "updater": updater,
            "force": bool,
            "timeout": int,
            "errorlog": str,
            "errorlog_level": loglevel,
            "errorlog_format": str,
            "statusfile": str,
            "pidfile": str,
        },
        "nsupdate": {
            "server": str,
            "key": str,
            "key_name": str,
            "key_algorithm": str,
        },
        "gandi": {
            "api_url": url,
            "api_key": str,
            "auth_api_url": url,
            "warning_api_key_expiration_delay": int,
        },
    }

    default = {
        "logging": {
            "level": "WARNING",
            "logfile": None,
            "logfile_format": "%(asctime)s - ee-spf-flattener - %(levelname)s - %(message)s",
            "logfile_level": "NOTSET",
            "console": False,
            "console_format": "%(asctime)s - %(levelname)s - %(message)s",
            "console_level": "NOTSET",
            "console_force_stderr": False,
        },
        "flattening": {
            "domain": None,
            "input_records": None,
            "name_servers": None,
            "spf_all_policy": DEFAULT_SPF_ALL_POLICY,
            "spf_record_name_format": DEFAULT_SPF_RECORD_NAME_FORMAT,
            "spf_record_max_length": DEFAULT_SPF_RECORD_MAX_LENGTH,
            "max_spf_records": DEFAULT_MAX_SPF_RECORDS,
            "dont_warn_about_spf_record_count": False,
            "ttl": DEFAULT_TTL,
        },
        "update": {
            "updater": DEFAULT_UPDATER,
            "force": None,
            "timeout": DEFAULT_TIMEOUT,
            "errorlog": False,
            "errorlog_level": "WARNING",
            "errorlog_format": "%(asctime)s - %(levelname)s - %(message)s",
            "pidfile": None,
            "statusfile": None,
        },
        "nsupdate": {
            "server": None,
            "key": None,
            "key_name": None,
            "key_algorithm": DEFAULT_NSUPDATE_KEY_ALGORITHM,
        },
        "gandi": {
            "api_url": DEFAULT_GANDI_API_URL,
            "api_key": None,
            "auth_api_url": DEFAULT_GANDI_AUTH_API_URL,
            "warning_api_key_expiration_delay": DEFAULT_GANDI_WARNING_API_KEY_EXPIRATION_DELAY,
        },
    }
    _resolver = None
    _logfile_log_handler = None
    _console_log_handler = None
    _update_errorlog_log_handler = None

    @property
    def args_parser(self):
        """Get arguments parser"""
        if self._args_parser:
            return self._args_parser
        self._args_parser = argparse.ArgumentParser(
            description="Script to compute flatten SPF record for a domain and dump it on STDOUT"
        )
        self._args_parser.add_argument(
            "-v",
            "--verbose",
            action="store_true",
            dest="logging_verbose",
            help="Enable verbose mode",
        )
        self._args_parser.add_argument(
            "-d", "--debug", action="store_true", dest="logging_debug", help="Enable debug mode"
        )
        self._args_parser.add_argument(
            "-l",
            "--log-file",
            action="store",
            type=str,
            dest="logging_logfile",
            help=f"Log file path (default: {self.default['logging']['logfile']})",
        )
        self._args_parser.add_argument(
            "--log-file-level",
            type=loglevel,
            dest="logging_logfile_level",
            help=(
                f"Log file level (if enabled, optional, "
                f"default: {self.default['logging']['logfile_level']}, "
                # pylint: disable=protected-access
                f"possible values: {', '.join(logging._nameToLevel)})"
            ),
        )
        self._args_parser.add_argument(
            "-C",
            "--console",
            action="store_true",
            dest="logging_console",
            help="Always log on console (even if log file is configured)",
        )
        self._args_parser.add_argument(
            "--console-level",
            type=loglevel,
            dest="logging_console_level",
            help=(
                f"Console log level (if enabled, optional, "
                f"default: {self.default['logging']['console_level']}, "
                # pylint: disable=protected-access
                f"possible values: {', '.join(logging._nameToLevel)})"
            ),
        )
        self._args_parser.add_argument(
            "--console-force-stderr",
            action="store_true",
            dest="logging_console_force_stderr",
            help=(
                f"Force console log on stderr (if enabled, optional, "
                f"default: {self.default['logging']['console_force_stderr']})"
            ),
        )
        self._args_parser.add_argument(
            "-j",
            "--just-try",
            action="store_true",
            dest="just_try",
            help="Enable just-try mode (no modification in Active Directory and OpenLDAP)",
        )
        self._args_parser.add_argument(
            "-c",
            "--config-file",
            help="Enable just-try mode (no modification in Active Directory and OpenLDAP)",
        )
        self._args_parser.add_argument(
            "-S",
            "--save-config",
            action="store_true",
            help="Save configuration in specified configuration file",
        )

        flat_opts = self._args_parser.add_argument_group("SPF flattening options")
        flat_opts.add_argument(
            "-D", "--domain", type=str, help="Domain name of the zone", dest="flattening_domain"
        )
        flat_opts.add_argument(
            "-i",
            "--input-records",
            type=input_records_arg,
            help=(
                "The input DNS record to use as source of SPF records: specified the record name "
                "and its type separated by colon (:). If type is omitted, assume it a TXT record. "
                "You can specify multiple input DNS records by using a comma separator (,). "
                "Supported record types are: "
                f"{', '.join([t.upper() for t in CRAWLER_SUPPORTED_RECORD_TYPES])}"
            ),
            dest="flattening_input_records",
        )
        flat_opts.add_argument(
            "-N",
            "--name-servers",
            help=(
                "DNS resolvers to use separated by comma (default: use system defined nameservers)"
            ),
            type=dns_resolvers,
            dest="flattening_name_servers",
        )
        flat_opts.add_argument(
            "-A",
            "--spf-all-policy",
            choices=list(SPF_ALL_POLICIES),
            help=(
                f"SPF all policy. Possible values: {', '.join(SPF_ALL_POLICIES)} "
                f"(default: {self.default['flattening']['spf_all_policy']})"
            ),
            dest="flattening_spf_all_policy",
        )
        flat_opts.add_argument(
            "-F",
            "--spf-record-name-format",
            type=spf_record_name_format,
            help=(
                "SPF record name format compute by replacing {} by the record index (default: "
                f"{self.default['flattening']['spf_record_name_format']})"
            ),
            dest="flattening_spf_record_name_format",
        )
        flat_opts.add_argument(
            "-M",
            "--spf-record-max-length",
            type=int,
            help=(
                "SPF record max length "
                f"(default: {self.default['flattening']['spf_record_max_length']})"
            ),
            dest="flattening_spf_record_max_length",
        )
        flat_opts.add_argument(
            "-m",
            "--max-spf-records",
            type=int,
            help=(
                "Maximum SPF records count "
                f"(default: {self.default['flattening']['max_spf_records']})"
            ),
            dest="flattening_max_spf_records",
        )
        flat_opts.add_argument(
            "--dont-warn-about-spf-record-count",
            action="store_true",
            help="Disable warnings about SPF records count",
            dest="flattening_dont_warn_about_spf_record_count",
        )
        flat_opts.add_argument(
            "--ttl",
            type=int,
            help=f"SPF records TTL (in seconds, default: {self.default['flattening']['ttl']})",
            dest="flattening_ttl",
        )

        self._args_parser.set_defaults(subcommand=None)
        subparsers = self._args_parser.add_subparsers(dest="subcommand", help="sub-command help")

        subparsers.add_parser("dump", help="Compute and dump flatten SPF records on STDOUT")

        update_parser = subparsers.add_parser(
            "update",
            help="Compute and update flatten SPF records",
        )

        update_parser.add_argument(
            "--updater",
            type=updater,
            choices=list_updaters(),
            help=f"SPF records updater (default: {self.default['flattening']['ttl']})",
            dest="updater",
        )

        update_parser.add_argument(
            "-f",
            "--force-update",
            action="store_true",
            help="Force to send NS update query even if flatten SPF records are up-to-date",
            dest="update_force",
        )
        update_parser.add_argument(
            "-t",
            "--update-timeout",
            type=int,
            help=f"Update query timeout (in seconds, default: {DEFAULT_TIMEOUT})",
            dest="update_timeout",
        )
        update_parser.add_argument(
            "-e",
            "--errorlog",
            action="store",
            type=str,
            dest="update_errorlog",
            help=f"Error log file path (default: {self.default['update']['errorlog']})",
        )
        update_parser.add_argument(
            "--errorlog-level",
            type=loglevel,
            dest="update_errorlog_level",
            help=(
                f"Error log file level (if enabled, optional, "
                f"default: {self.default['update']['errorlog_level']}, "
                # pylint: disable=protected-access
                f"possible values: {', '.join(logging._nameToLevel)})"
            ),
        )
        update_parser.add_argument(
            "-P",
            "--pid-file",
            action="store",
            type=str,
            dest="update_pidfile",
            help=(
                "PID file path: store the PID of the current process and do not start another if "
                f"already running (default: {self.default['update']['pidfile']})"
            ),
        )
        update_parser.add_argument(
            "-S",
            "--update-status-file",
            action="store",
            type=str,
            dest="update_statusfile",
            help=(
                "Status file path: touch after each successful update "
                f"(default: {self.default['update']['statusfile']})"
            ),
        )

        nsupdate_opts = update_parser.add_argument_group("NSupdate SPF records updater options")

        nsupdate_opts.add_argument(
            "-N",
            "--nsupdate-server",
            type=str,
            help="The primary nameserver of the domain to send NS update query",
            dest="nsupdate_server",
        )
        nsupdate_opts.add_argument(
            "-k",
            "--nsupdate-key",
            type=str,
            help="The base64 encoded key to sign the NS update query",
            dest="nsupdate_key",
        )
        nsupdate_opts.add_argument(
            "-K",
            "--nsupdate-key-name",
            type=str,
            help="The key name of the specified key to sign the NS update query",
            dest="nsupdate_key_name",
        )
        nsupdate_opts.add_argument(
            "-A",
            "--nsupdate-key-algorithm",
            type=str,
            help=f"The key algorithm (default: {DEFAULT_NSUPDATE_KEY_ALGORITHM})",
            dest="nsupdate_key_algorithm",
        )

        gandi_opts = update_parser.add_argument_group("Gandi LiveDNS SPF records updater options")

        gandi_opts.add_argument(
            "--gandi-api-url",
            type=str,
            help=f"Gandi API URL (default: {DEFAULT_GANDI_API_URL})",
            dest="gandi_api_url",
        )
        gandi_opts.add_argument(
            "--gandi-api-key",
            type=str,
            help="The Gandi API key",
            dest="gandi_api_key",
        )

        gandi_opts.add_argument(
            "--gandi-auth-api-url",
            type=str,
            help=f"Gandi authentication API URL (default: {DEFAULT_GANDI_AUTH_API_URL})",
            dest="gandi_auth_api_url",
        )

        gandi_opts.add_argument(
            "--gandi-warning-api-key-expiration-delay",
            type=int,
            help=(
                "Gandi authentication API URL (in days, set to None to disable, "
                f"default: {DEFAULT_GANDI_WARNING_API_KEY_EXPIRATION_DELAY})"
            ),
            dest="gandi_warning_api_key_expiration_delay",
        )

        subparsers.add_parser("check_updater", help="Check updater")

        return self._args_parser

    def parse_args(self):
        """Parse CLI arguments"""
        if self.args:
            return self.args.subcommand
        # Parse arguments
        self.args = self.args_parser.parse_args()
        self.init_logging()

        if self.args.config_file:
            self.load_config_file(self.args.config_file, allow_not_found=self.args.save_config)
            if self.args.save_config:
                self.save_config_file(self.args.config_file)
        elif self.args.save_config:
            self.error("You must provide the configuration file path to save it.")

        if self.config_path:
            self.init_logging()
            LOG.info("Configuration loadded from %s", self.config_path)

        if self.args.subcommand == "update" and self.get("update.pidfile"):
            pidfile = self.get("update.pidfile")
            if os.path.exists(pidfile):
                with open(pidfile, encoding="utf8") as pidfile_fd:
                    pid = pidfile_fd.read()
                self.error("Another update process with PID %s is already running, stop", pid)
            LOG.debug("PID file %s not present, install it", pidfile)
            with open(pidfile, "w", encoding="utf8") as pidfile_fd:
                pidfile_fd.write(str(os.getpid()))
            LOG.debug("PID file %s installed, configured removing hook at exit", pidfile)
            atexit.register(self._remove_update_pidfile)

        return self.args.subcommand

    def _remove_update_pidfile(self):
        """Remove update PID file"""
        if self.args.subcommand != "update" or not self.get("update.pidfile"):
            return
        pidfile = self.get("update.pidfile")
        LOG.debug("Remove PID file %s", pidfile)
        os.remove(pidfile)

    def init_logging(self):
        """Initialize logging from configuration"""
        # Initialize logging
        root_logger = logging.getLogger()

        # Remove all previously configured handlers (and default ones)
        root_logger.handlers = []

        # Set root logger level
        if self.get("logging.debug"):
            root_logger.setLevel(logging.DEBUG)
        elif self.get("logging.verbose"):
            root_logger.setLevel(logging.INFO)
        else:
            root_logger.setLevel(
                logging._nameToLevel[self.get("logging.level")]  # pylint: disable=protected-access
            )

        # Handle logfile handler
        if self.get("logging.logfile"):
            self._logfile_log_handler = logging.FileHandler(self.get("logging.logfile"))
            self._logfile_log_handler.setLevel(
                # pylint: disable=protected-access
                logging._nameToLevel[self.get("logging.logfile_level")]
            )
            self._logfile_log_handler.setFormatter(
                logging.Formatter(self.get("logging.logfile_format"))
            )
            root_logger.addHandler(self._logfile_log_handler)

        # Handle console handler
        if not self.get("logging.logfile") or self.get("logging.console"):
            # pylint: disable=protected-access
            console_level = logging._nameToLevel[self.get("logging.console_level")]
            console_formater = logging.Formatter(self.get("logging.console_format"))

            if console_level < logging.WARNING:
                stdout_console_handler = logging.StreamHandler(
                    sys.stderr if self.get("logging.console_force_stderr") else sys.stdout
                )
                stdout_console_handler.addFilter(StdoutInfoFilter())
                stdout_console_handler.setLevel(console_level)
                stdout_console_handler.setFormatter(console_formater)
                root_logger.addHandler(stdout_console_handler)

            stderr_console_handler = logging.StreamHandler(sys.stderr)
            stderr_console_handler.setLevel(
                console_level if console_level > logging.WARNING else logging.WARNING
            )
            stderr_console_handler.setFormatter(console_formater)
            root_logger.addHandler(stderr_console_handler)

        # Handle update errorlog file handler
        if self.args and self.args.subcommand == "update" and self.get("update.errorlog"):
            self._update_errorlog_log_handler = logging.FileHandler(
                self.get("update.errorlog"),
                delay=True,  # Create file only on first error
            )
            self._update_errorlog_log_handler.setLevel(
                # pylint: disable=protected-access
                logging._nameToLevel[self.get("update.errorlog_level")]
            )
            self._update_errorlog_log_handler.setFormatter(
                logging.Formatter(self.get("update.errorlog_format"))
            )
            root_logger.addHandler(self._update_errorlog_log_handler)

    def load_config_file(self, path, allow_not_found=False):
        """Load configuration file"""
        try:
            with open(path, encoding="utf8") as fd:
                raw_config_data = yaml.safe_load(fd)
        except FileNotFoundError:
            if allow_not_found:
                return
            raise
        config_data = {}
        for section, section_config in self.default.items():
            if section not in raw_config_data:
                continue
            if section not in config_data:
                config_data[section] = {}
            for key in section_config:
                if key not in raw_config_data[section] or raw_config_data[section][key] is None:
                    continue
                config_data[section][key] = raw_config_data[section][key]
        self.config_data = config_data
        self.config_path = path

    def save_config_file(self, path, new_config_data=None):
        """Save configuration file"""
        if not new_config_data:
            new_config_data = {}
            for section, section_config in self.default.items():
                new_config_data[section] = {}
                for key in section_config:
                    new_config_data[section][key] = self.get(f"{section}.{key}")
        with open(path, "w", encoding="utf8") as fd:
            yaml.dump(new_config_data, fd, default_flow_style=False)

    def get(self, key, default=None, required=False, config_data=None):
        """Get configuration key"""
        value = self._get_arg(key)
        LOG.debug("config.get(%s): value from args = '%s'", key, value)
        if value is None:
            value = self._get_config(key, default=default, config_data=config_data)
            LOG.debug("config.get(%s): value from config = '%s'", key, value)
        if value is None and required:
            self.error(f"Configuration parameter {key} is missing.")
        value = self.replace_variables(value, config_data=config_data)
        return value

    def _get_arg(self, key):
        """Get configuration parameter from CLI arguments"""
        arg_key = key.replace(".", "_")
        if self.args and hasattr(self.args, arg_key) and getattr(self.args, arg_key):
            return getattr(self.args, arg_key)
        return None

    def _get_config(self, key, default=None, config_data=None):
        """Get configuration parameter from configuration file"""
        if config_data is None:
            default = default if default else self._get_config(key, config_data=self.default)
        splitted_keys = key.split(".")
        value = config_data if config_data else self.config_data
        value_type = self.config_types
        for k in splitted_keys:
            if not isinstance(value, dict) or k not in value:
                return default
            value = value[k]
            value_type = value_type.get(k, {})
        if callable(value_type) and value is not None:
            LOG.debug("_get_config(%s): cast found value with %s", key, value_type)
            try:
                value = value_type(value)
            except ValueError:
                self.error(
                    "Invalid configuration parameter %s (%s)",
                    key,
                    value,
                )
        return value

    def replace_variables(self, value, config_data=None):
        """Replace variable in specified value"""
        if isinstance(value, list):
            return [self.replace_variables(v, config_data=config_data) for v in value]
        if not isinstance(value, str):
            return value
        orig_value = value
        for pattern, key in re.findall(r"(\$\{([^\}]+)\})", value):
            replacement_value = self.get(key, required=True, config_data=config_data)
            LOG.debug(
                "replace_variables(%s): found pattern '%s' (%s) => replace by '%s'",
                orig_value,
                pattern,
                key,
                replacement_value,
            )
            value = value.replace(pattern, replacement_value)
        LOG.debug("replace_variables(%s): final value = '%s'", orig_value, value)
        return value

    def error(self, message, *args):
        """Trigger an error and exit"""
        if args:
            message = message % args
        if self.args_parser:
            self.args_parser.error(message)
        print(message)
        sys.exit(1)

    @property
    def just_try(self):
        """Retrieve just-try mode status"""
        return self.args and getattr(self.args, "just_try")

    @property
    def resolver(self):
        """Retrieve DNS resolver configured with specified name servers"""
        if self._resolver:
            return self._resolver
        # Create DNS resolver to use during flattening process
        self._resolver = Resolver()
        name_servers = self.get("flattening.name_servers")
        if name_servers:
            LOG.debug("Use specified DNS resolvers (%s)", ", ".join(name_servers))
            self._resolver.nameservers = name_servers
        else:
            LOG.debug("Use system default DNS resolvers")
        return self._resolver


# pylint: disable=too-few-public-methods
class StdoutInfoFilter(logging.Filter):
    """Logging filter to keep messages only >= logging.INFO"""

    def filter(self, record):
        return record.levelno in (logging.DEBUG, logging.INFO)


config = Config()
