"""Crawler for SPF record"""
import logging

import dns
from netaddr import IPAddress

from ee_spf_flattener.config import config
from ee_spf_flattener.const import CRAWLER_SUPPORTED_RECORD_TYPES
from ee_spf_flattener.handlers import handler_mapping
from ee_spf_flattener.mechanisms import tokenize

LOG = logging.getLogger(__name__)


def spf2ips(records, domain, resolver=None):
    """Get IP addresses from SPF records"""
    resolver = resolver if resolver else config.resolver
    ips = set()
    for record in records:
        for ip in crawl(record["name"], record["type"], domain, resolver):
            ips.add(ip)
    return ips


def crawl(rrname, rrtype, domain, resolver=None):
    """Crawl IP addresses for a specified SPF record"""
    assert (
        rrtype in CRAWLER_SUPPORTED_RECORD_TYPES
    ), f"No crawler implement for {rrtype.upper()} records"
    return globals()[f"crawl_{rrtype}"](rrname, domain, resolver)


def crawl_txt(rrname, domain, resolver=None):
    """Crawl SPF IP address from specified SPF record name"""
    resolver = resolver if resolver else config.resolver
    try:
        LOG.debug("Looking for SPF entries by resolving TXT records for %s", rrname)
        answers = resolver.query(dns.name.from_text(rrname), "TXT")
    except dns.resolver.NXDOMAIN:
        LOG.error("Fail to resolve TXT record for %s: does not exists", rrname)
    except Exception:  # pylint: disable=broad-exception-caught
        LOG.exception("Fail to resolve TXT records for %s", rrname)
    else:
        LOG.debug("Answers found:\n- %s", "\n- ".join(map(str, answers)))
        for answer in map(str, answers):
            if not answer.startswith('"v=spf1 '):
                LOG.debug("Answer is not an SPF record, ignore it (%s)", answer)
                continue
            for pair in tokenize(answer):
                rname, rtype = pair
                if rtype is None:
                    continue
                if rtype == "txt":
                    yield from crawl(rname, "txt", domain, resolver)
                    continue
                try:
                    yield from handler_mapping[rtype](rname, domain, resolver)
                except (dns.resolver.NXDOMAIN, dns.resolver.NoAnswer) as e:
                    LOG.warning(e)


def crawl_mx(rrname, domain, resolver=None):  # pylint: disable=unused-argument
    """Crawl SPF IP address from specified MX record name"""
    resolver = resolver if resolver else config.resolver
    try:
        LOG.debug("Looking for SPF entries from MX records of %s", rrname)
        answers = resolver.query(dns.name.from_text(rrname), "MX")
    except dns.resolver.NXDOMAIN:
        LOG.error("Fail to resolve MX records for %s: does not exists", rrname)
    except Exception:  # pylint: disable=broad-exception-caught
        LOG.exception("Fail to resolve MX records for %s", rrname)
    else:
        for mailexchange in answers:
            try:
                for ip in resolver.query(mailexchange.exchange, "a"):
                    yield IPAddress(ip.address)
            except (dns.resolver.NXDOMAIN, dns.resolver.NoAnswer) as e:
                LOG.warning(e)


def crawl_ip(rrname, rrtype, domain, resolver=None):  # pylint: disable=unused-argument
    """Crawl SPF IP address from specified record name"""
    resolver = resolver if resolver else config.resolver
    try:
        LOG.debug("Looking for SPF entries from %s records of %s", rrtype.upper(), rrname)
        answers = resolver.query(dns.name.from_text(rrname), rrtype)
    except (dns.resolver.NXDOMAIN, dns.resolver.NoAnswer):
        LOG.error("Fail to resolve %s records for %s: does not exists", rrtype.upper(), rrname)
    except Exception:  # pylint: disable=broad-exception-caught
        LOG.exception("Fail to resolve %s records for %s", rrtype.upper(), rrname)
    else:
        for answer in answers:
            if rrtype in ["cname", "ns"]:
                try:
                    LOG.debug(
                        "Looking for IP address for %s %s record target %s",
                        rrname,
                        rrtype.upper(),
                        answer.target,
                    )
                    for ip in resolver.query(answer.target):
                        yield IPAddress(ip.address)
                except (dns.resolver.NXDOMAIN, dns.resolver.NoAnswer) as e:
                    LOG.warning(e)
            else:
                yield IPAddress(answer.address)


def crawl_a(rrname, domain, resolver=None):
    """Crawl SPF IP address from specified A record name"""
    return crawl_ip(rrname, "a", domain, resolver=resolver)


def crawl_aaaa(rrname, domain, resolver=None):
    """Crawl SPF IP address from specified AAAA record name"""
    return crawl_ip(rrname, "aaaa", domain, resolver=resolver)


def crawl_cname(rrname, domain, resolver=None):
    """Crawl SPF IP address from specified CNAME record name"""
    return crawl_ip(rrname, "cname", domain, resolver=resolver)


def crawl_ns(rrname, domain, resolver=None):
    """Crawl SPF IP address from specified NS record name"""
    return crawl_ip(rrname, "ns", domain, resolver=resolver)
