"""Formatting SPF records helpers"""

from netaddr import AddrFormatError, IPNetwork, IPSet

from ee_spf_flattener.config import config
from ee_spf_flattener.const import SPF_ALL_POLICIES


def compute_spf_records(
    ips, domain=None, all_policy=None, record_name_format=None, max_length=None
):
    """
    Compute SPF records form address IP list to included as a dict
    Note: Length of TXT record is limited, we must split result in multiple SPF records
    """
    records = {}
    base_record = "v=spf1"
    domain = domain if domain else config.get("flattening.domain", required=True)
    all_policy = SPF_ALL_POLICIES[
        all_policy if all_policy else config.get("flattening.spf_all_policy")
    ]
    record_name_format = (
        record_name_format
        if record_name_format
        else config.get("flattening.spf_record_name_format")
    )
    max_length = max_length if max_length else config.get("flattening.spf_record_max_length")

    current_record = base_record
    current_record_name = record_name_format.format("")
    next_record_name = record_name_format.format(1)

    next_record_format = f" include:{{}} {all_policy}"
    for ip in ips:
        ip = f" {ip}"
        if (
            len(current_record) + len(ip) + len(next_record_format.format(next_record_name))
            > max_length
        ):
            current_record += next_record_format.format(next_record_name)
            records[current_record_name] = current_record
            current_record = base_record
            current_record_name = record_name_format.format(len(records))
            next_record_name = record_name_format.format(len(records) + 1)
        current_record += ip
    current_record += f" {all_policy}"
    records[current_record_name] = current_record
    return records


def format_spf_record(name, value, ttl=None):
    """Format an SPF record"""
    # pylint: disable=consider-using-f-string
    return f'{name}{" {}".format(ttl) if ttl else ""} IN TXT "{value}"'


def ips_to_spf_strings(ips):
    """Convert IP list to SPF entry as string"""
    other_tokens = []
    for ip in ips:
        try:
            IPNetwork(ip)
        except AddrFormatError:
            other_tokens.append(ip)
    for token in other_tokens:
        ips.remove(token)
    ips = [str(i) for i in IPSet(ips).iter_cidrs()]
    ips = ["ip6:" + ip if ":" in ip else "ip4:" + ip.replace("/32", "") for ip in ips]
    return ips + other_tokens
