"""SPF entry handlers"""

from dns.name import from_text
from netaddr import IPAddress, IPNetwork


def handle_ip(name, domain, resolver):  # pylint: disable=unused-argument
    """Handler for SPF IP entry"""
    yield name


def handle_mx(name, domain, resolver):  # pylint: disable=unused-argument
    """Handler for SPF MX entry"""
    answers = resolver.query(from_text(domain), "mx")
    for mailexchange in answers:
        ips = resolver.query(mailexchange.exchange, "a")
        for ip in ips:
            yield IPAddress(ip.address)


def handle_mx_domain(name, domain, resolver):  # pylint: disable=unused-argument
    """Handler for SPF MX domain entry"""
    answers = resolver.query(from_text(name), "mx")
    for mailexchange in answers:
        ips = resolver.query(mailexchange, "a")
        for ip in ips:
            yield IPAddress(ip.address)


def handle_mx_prefix(name, domain, resolver):  # pylint: disable=unused-argument
    """Handler for SPF MX prefix entry"""
    _name, prefix = name
    answers = resolver.query(from_text(domain), "mx")
    for mailexchange in answers:
        ips = resolver.query(mailexchange.exchange, "a")
        for ip in ips:
            yield IPNetwork(f"{ip}/{prefix}")


def handle_mx_domain_prefix(name, domain, resolver):  # pylint: disable=unused-argument
    """Handler for SPF MX domain prefix entry"""
    _name, prefix = name
    answers = resolver.query(from_text(_name), "mx")
    for mailexchange in answers:
        ips = resolver.query(mailexchange, "a")
        for ip in ips:
            yield IPNetwork(f"{ip}/{prefix}")


def handle_a(name, domain, resolver):  # pylint: disable=unused-argument
    """Handler for SPF A entry"""
    answers = resolver.query(from_text(domain), "a")
    for ip in answers:
        yield IPAddress(ip.address)


def handle_a_domain(name, domain, resolver):  # pylint: disable=unused-argument
    """Handler for SPF A domain entry"""
    answers = resolver.query(from_text(name), "a")
    for ip in answers:
        yield IPAddress(ip.address)


def handle_a_prefix(name, domain, resolver):
    """Handler for SPF A prefix entry"""
    _name, prefix = name
    answers = resolver.query(from_text(domain), "a")
    for ip in answers:
        yield IPNetwork(f"{ip}/{prefix}")


def handle_a_domain_prefix(name, domain, resolver):  # pylint: disable=unused-argument
    """Handler for SPF A domain prefix entry"""
    _name, prefix = name
    answers = resolver.query(from_text(_name), "a")
    for ip in answers:
        yield IPNetwork(f"{ip}/{prefix}")


def handle_ptr(name, domain, resolver):  # pylint: disable=unused-argument
    """Handler for SPF PTR entry"""
    yield f"ptr:{name}"


def handle_exists(name, domain, resolver):  # pylint: disable=unused-argument
    """Handler for SPF exists entry"""
    yield f"exists:{name}"


handler_mapping = {
    "ip": handle_ip,
    "mx": handle_mx,
    "mx_prefix": handle_mx_prefix,
    "mx_domain": handle_mx_domain,
    "mx_domain_prefix": handle_mx_domain_prefix,
    "a": handle_a,
    "a_domain": handle_a_domain,
    "a_prefix": handle_a_prefix,
    "a_domain_prefix": handle_a_domain_prefix,
    "ptr": handle_ptr,
    "exists": handle_exists,
}
