"""Gandi SPF records updater"""
import datetime
import json
import logging

import requests
from requests.exceptions import HTTPError

try:
    from requests.exceptions import JSONDecodeError
except ImportError:
    from json.decoder import JSONDecodeError

from ee_spf_flattener.config import config
from ee_spf_flattener.updater import BaseUpdater

LOG = logging.getLogger(__name__)


class Updater(BaseUpdater):
    """SPF records updater"""

    api_url = None
    api_key = None
    auth_api_url = None
    warning_api_key_expiration_delay = None

    def initialize(self):
        """Initialize updater"""
        if not self.api_url:
            self.api_url = config.get("gandi.api_url", required=True)
            self.api_key = config.get("gandi.api_key", required=True)
            self.auth_api_url = config.get("gandi.auth_api_url", required=True)
            self.warning_api_key_expiration_delay = config.get(
                "gandi.warning_api_key_expiration_delay"
            )
            self._check_api_key_expiration()

    def __call_api(
        self,
        suburl,
        method=None,
        data=None,
        headers=None,
        auth_api=False,
        decode_json=True,
        return_error=False,
        log_error=True,
    ):
        """Call Gandi API"""
        url = f"{self.api_url if not auth_api else self.auth_api_url}/{suburl}"
        method = method if method else "GET"
        headers = headers if headers else {}
        headers["authorization"] = f"Bearer {self.api_key}"

        response = requests.request(
            method,
            url,
            headers=headers,
            data=json.dumps(data) if data else None,
            timeout=config.get("update.timeout"),
        )
        try:
            response.raise_for_status()
            return response.json() if decode_json else response.text
        except JSONDecodeError as err:
            LOG.error(
                "Error occurred decoding Gandi API JSON result on %s request to %s: %s\n%s",
                method,
                url,
                err,
                response.text,
            )
        except HTTPError as err:
            if return_error:
                return response
            if log_error:
                LOG.error(
                    "Error occurred calling Gandi API on %s request to %s:%s\n%s",
                    method,
                    url,
                    err,
                    response.text,
                )
        return False

    @staticmethod
    def record_url(record, record_type=None):
        """Compute one record URL"""
        record_name = record.split(".")[0]
        domain = ".".join(record.split(".")[1:])
        url = f"livedns/domains/{domain}/records/{record_name}"
        if record_type:
            url += f"/{record_type}"
        return url

    def set_record(self, record, record_type, value, ttl=None):
        """Set the value of one record (replace all existing values)"""
        response = self.__call_api(
            self.record_url(record),
            method="PUT",
            data={
                "items": [
                    {
                        "rrset_type": record_type,
                        "rrset_values": [value],
                        "rrset_ttl": ttl if ttl is not None else config.get("flattening.ttl"),
                    }
                ],
            },
            headers={"content-type": "application/json"},
        )
        return bool(response)

    def delete_record(self, record):
        """Delete all values of one record"""
        response = self.__call_api(
            self.record_url(record),
            method="DELETE",
            decode_json=False,
            return_error=True,
        )
        if isinstance(response, HTTPError):
            # Error returned: allow only 404 error on not found DNS record
            try:
                response = response.json()
                return (
                    response.response.get("object") == "dns-record"
                    and response.get("response") == 404
                    and response.get("cause") == "Not Found"
                )
            except JSONDecodeError as err:
                LOG.error(
                    "Error occurred decoding Gandi API JSON error on deleting record %s: %s\n%s",
                    record,
                    err,
                    response.response.text,
                )
                return False
        elif isinstance(response, str):
            # Empty string returned on success
            return response == ""
        return False

    def record_exists(self, record, record_type):
        """Check if one record exists"""
        response = self.__call_api(
            self.record_url(record, record_type=record_type), log_error=False
        )
        return bool(response)

    def update_spf_records(self, records):
        """Update SPF records by send an NS update query"""
        error = False
        for name, value in records.items():
            if self.set_record(name, "TXT", f'"{value}"'):
                LOG.debug("Record %s updated", name)
            else:
                LOG.error("Fail to update record %s with value '%s'", name, value)
                error = True

        LOG.info("Lookup for old SPF record need to be cleaned")
        count = 0
        while True:
            name = self.get_record_name(count)
            count += 1
            if name in records:
                continue
            if not self.record_exists(name, record_type="TXT"):
                # No more old SPF record to clean, stop
                break
            LOG.info("An old %s SPF record detected, purge it", name)
            if self.delete_record(name):
                LOG.debug("Record %s deleted", name)
            else:
                LOG.error("Fail to delete record %s", name)
                error = True
        if error:
            LOG.info("Some error occurred updating SPF records")
            return False
        LOG.info("SPF records updated")
        return True

    def _check_api_key_expiration(self):
        """Check API key expiration"""
        if not self.warning_api_key_expiration_delay:
            return
        try:
            response = self.__call_api("tokeninfo", auth_api=True)
            expiration_delay = datetime.timedelta(seconds=response["expires_in"])
            expiration_date = datetime.datetime.now() + expiration_delay
        except Exception:  # pylint: disable=broad-exception-caught
            LOG.warning("Fail to retrieve Gandi API key expiration date")
            return
        LOG.log(
            logging.WARNING
            if expiration_delay.days < self.warning_api_key_expiration_delay
            else logging.DEBUG,
            "Gandi API key will expire in %d day(s) (%s)",
            expiration_delay.days,
            expiration_date.strftime("%Y/%m/%d, %T"),
        )
