"""NSupdate SPF records updater"""
import logging

import dns
import dns.tsigkeyring
import dns.update
from netaddr import IPAddress
from netaddr.core import AddrFormatError

from ee_spf_flattener.config import config
from ee_spf_flattener.updater import BaseUpdater

LOG = logging.getLogger(__name__)


class Updater(BaseUpdater):
    """SPF records updater"""

    ns_server = None
    ns_key_name = None
    ns_key = None
    ns_key_algorithm = None

    def initialize(self):
        """Initialize updater"""
        self.ns_server = config.get("nsupdate.server", required=True)
        self.ns_key_name = config.get("nsupdate.key_name", required=True)
        self.ns_key = config.get("nsupdate.key", required=True)
        self.ns_key_algorithm = config.get("nsupdate.key_algorithm")

    def update_spf_records(self, records):
        """Update SPF records by send an NS update query"""
        # Handle optional parameters

        # Ensure we have the IP address of the NS server
        ns_server_ips = self.hostname2ips(self.ns_server)
        if not ns_server_ips:
            LOG.error("Fail to resolve NS server %s IP address", self.ns_server)
            return False

        # Prepare the payload for DNS record update in the given zone/domain
        update = dns.update.Update(
            dns.name.from_text(self.get_records_domain(records)),
            keyring=dns.tsigkeyring.from_text({self.ns_key_name: self.ns_key}),
            keyname=self.ns_key_name,
            keyalgorithm=dns.name.from_text(self.ns_key_algorithm),
        )

        for name, value in records.items():
            update.replace(
                dns.name.from_text(name), config.get("flattening.ttl"), "TXT", f'"{value}"'
            )

        LOG.info("Lookup for old SPF record need to be cleaned")

        ns_server_resolver = dns.resolver.Resolver()
        ns_server_resolver.nameservers = ns_server_ips

        count = 0
        while True:
            name = self.get_record_name(count)
            count += 1
            if name in records:
                continue
            try:
                ns_server_resolver.query(dns.name.from_text(name), "TXT")
            except dns.resolver.NXDOMAIN:
                # No more old SPF record to clean, stop
                break
            except Exception:  # pylint: disable=broad-exception-caught
                LOG.exception(
                    "Error occurred resolving %s TXT record, stop old SPF records cleaning", name
                )
                break
            LOG.info("An old %s SPF record detected, purge it", name)
            update.delete(dns.name.from_text(name), "TXT")

        # Submit the update query to the DNS server to apply change
        LOG.debug("NS update query:\n%s", update)
        try:
            response = dns.query.tcp(update, ns_server_ips[0], timeout=config.get("update.timeout"))
        except dns.tsig.PeerBadKey:
            LOG.error("Provided key refused by the NS server")
            return False
        LOG.info("SPF records updated")
        LOG.debug("Update result: %s", response)
        return True

    @staticmethod
    def hostname2ips(value, resolver=None):
        """Retrieve IP addresses corresponding to specified hostname"""
        try:
            IPAddress(value)
            return [value]
        except AddrFormatError:
            pass

        resolver = resolver if resolver else config.resolver
        try:
            answers = [str(answer) for answer in resolver.query(dns.name.from_text(value))]
            LOG.debug("Name %s resolved as %s", value, ", ".join(answers))
            return answers
        except (dns.resolver.NXDOMAIN, dns.resolver.NoAnswer) as e:
            LOG.error("DNS name %s seem does not exists (%s)", value, e)
        return False
